/*
 * Copyright (C) 2013 Texas Instruments Incorporated - http://www.ti.com/
 *
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "mpm_transport_hyplnk_interface.h"
#include "uioutils.h"
#include "workaround.c"

extern hyplnk_peripheral hyplnkActiveIF[NUM_INTERFACES];
/***********************************************************************/
int sl1500_get_nodeid(void)
{
        char st[MAX_LINE_LENGTH];
        FILE *fp = fopen("/proc/device-tree/sl1500/node/id", "r");
        if (fp==NULL) return -1;
        fscanf(fp, "%s", st);
        if (fp) fclose(fp);
        return atoi(st);
}
static void timpi_dbg_serdes_dump(int node_id, int chkpoint, uint32_t map_base, uint32_t hyp_base) {
unsigned int i, j, addr;

  mpm_printf (1, "\n#[%d]# DUMP REGISTERS, chkpoint %d serdes_base %08x hyplnk_base %08x\n", 
    node_id, chkpoint, map_base, hyp_base);
  mpm_printf (1, "\n#[%d]# DUMP Hyplnk REGISTERS (0x00-0x5c), chkpoint %d\n", node_id, chkpoint); fflush(stdout);
  for (i = 0; i < 24; i += 4)
  {
    addr = 0x0000 + i * 4;
    mpm_printf (1, "\n#[%d]# %03d Hyplnk (%04x)=%08x %08x %08x %08x)",
             node_id, chkpoint, addr,
             *((unsigned int *)(hyp_base + 0x0000 + i * 4)),
             *((unsigned int *)(hyp_base + 0x0004 + i * 4)),
             *((unsigned int *)(hyp_base + 0x0008 + i * 4)),
             *((unsigned int *)(hyp_base + 0x000c + i * 4))
           );
  }
#if 0
  for (i = 0; i < 64; i ++)
  {
    addr = 0x0000 + i * 4;
    printf ("\n[%d] %03d CMU0 SERDES(%04x)=%08x)",
             node_id, chkpoint, addr,
             *((unsigned int *)(map_base + 0x0000 + i * 4)));
  }
  for (j = 0; j < 4; j ++)
  {
    for (i = 0; i < 128; i ++)
    {
      addr = 0x200 + 0x200 * j + i * 4;
      printf ("\n[%d] %03d LANE%d SERDES(%04x)=%08x)",
         node_id, chkpoint, j, addr,
        *((unsigned int *)(map_base + addr)));
    }
  }
  for (i = 0; i < 128; i ++)
  {
    addr = 0x0A00 + i * 4;
    printf ("\n[%d] %03d COMLANE SERDES(%04x)=%08x)",
             node_id, chkpoint, addr,
             *((unsigned int *)(map_base + addr)));
  }
#endif
  mpm_printf (1, "\n#[%d]# DUMP SERDES REGISTERS, chkpoint %d\n", node_id, chkpoint);
  for (i = 0; i < 16; i += 4)
  {
     addr = 0x1FC0 + i * 4;
     mpm_printf (1, "\n#[%d]# %03d WIZ SERDES(%04x)=%08x %08x %08x %08x)",
             node_id, chkpoint, addr,
             *((unsigned int *)(map_base + addr + 0x00)),
             *((unsigned int *)(map_base + addr + 0x04)),
             *((unsigned int *)(map_base + addr + 0x08)),
             *((unsigned int *)(map_base + addr + 0x0c))
           );
  }
}                                                                                                           
/***********************************************************************/
void * internal_hyplnk_mmap(uint32_t addr, uint32_t size, int device_fd)
{
    void * virt_addr;
    uint32_t page_size, mapaddr, offset;
    page_size = getpagesize();
    mapaddr = addr & (~(page_size - 1));
    offset = addr - mapaddr;
    virt_addr = offset + mmap(0, size + offset, (PROT_READ|PROT_WRITE), MAP_SHARED, device_fd, (off_t)mapaddr);
    if ((uint32_t)virt_addr == -1)
    {
            mpm_printf(1, "mmap failed!\n");
            return 0;
    }
    return virt_addr;	
}

void hyplnkPSCSetup(int port)
{
    int ret;
    char class_name[MAX_FILE_NAME_LENGTH];
    char dev_name[MAX_NAME_LENGTH];
    if (port == 0)
      ret = uioutil_get_peripheral_device("hyperlink0", class_name, 32);
    else
      ret = uioutil_get_peripheral_device("hyperlink1", class_name, 32);
    if ( ret < 0 ) {
      mpm_printf(1, "\n Error getting device ");
    }
    snprintf(dev_name, MAX_NAME_LENGTH, "/dev/%s", class_name);

    hyplnkActiveIF[port].uio_fd = open (dev_name, (O_RDWR | O_SYNC));
    if (hyplnkActiveIF[port].uio_fd < 0)
      mpm_printf(1, "\n Error opening device ");
    
} /* hyplnkPSCSetup */

void hyplnkPSCDisable(int port)
{
    close(hyplnkActiveIF[port].uio_fd);
} /* hyplnkPSCDisable */

void hyplnkSerdesDisable(mpm_transport_hyplnk_t *hypCfg)
{
    uint32_t baseAddr, i, serdes_regs = 0;
    int index;
    int iAtt, iBoost;
    for (index=0; index<2; index++) {
      if (hypCfg->fd[index] == -1) continue;
      if (index == 0)
          serdes_regs = CSL_HYPERLINK_0_SERDES_CFG_REGS;
      else if (index == 1)
          serdes_regs = CSL_HYPERLINK_1_SERDES_CFG_REGS;
      baseAddr = (uint32_t) internal_hyplnk_mmap(serdes_regs,0x2000, hypCfg->fd[index]);
      CSL_HyperlinkSerdesShutdown(baseAddr);
      CSL_SerDes_CMU_Restore_Default(baseAddr);
      CSL_SerDes_COMLANE_Restore_Default(baseAddr);
      for(i=0; i < hypCfg->hyplnkNumLanes; i++)
      {
          CSL_SerDes_Lane_Restore_Default(baseAddr, i);
      }
    }
}

void mpm_transport_hyplnk_startup_sequence(mpm_transport_hyplnk_t *hypCfg, int index)
{
    uint32_t hyplnk_base, serdes_base;
    int ret=0, loopback, i=0, j = 0, iTmpRevID, iTmp, iTmpECC, iTmpStat, attemptCNT;
    if (hypCfg->hyplnkDir == hyplnk_LOOPBACK) loopback = 1;
    else loopback = 0;
    if (index == 0) {
        serdes_base = (uint32_t) internal_hyplnk_mmap(CSL_HYPERLINK_0_SERDES_CFG_REGS,0x2000, hypCfg->fd[0]);
        hyplnk_base = (uint32_t) internal_hyplnk_mmap(CSL_HYPERLINK_0_SLV_CFG_REGS, 0x100, hypCfg->fd[0]);
    }
    else {
        serdes_base = (uint32_t) internal_hyplnk_mmap(CSL_HYPERLINK_1_SERDES_CFG_REGS,0x2000, hypCfg->fd[1]);
        hyplnk_base = (uint32_t) internal_hyplnk_mmap(CSL_HYPERLINK_1_SLV_CFG_REGS, 0x100, hypCfg->fd[1]);
    }
    if (!hyplnkPeripheralSerdesHealth(hyplnk_base, serdes_base)) goto startup_end;
    hyplnk_startupTX(serdes_base, hyplnk_base, loopback, index);

startup_end:
    munmap((void *)serdes_base, 0x2000);
    munmap((void *)hyplnk_base, 0x100);

    return ret;
}

void hyplnk_clear_int(uint32_t hyplnk_base) {
    uint32_t iTmp;
    int soc_node_id = sl1500_get_nodeid();

    // Interrupt Priority / Clear Register
    iTmp = *(uint32_t *) (hyplnk_base + 0x0c);
    if (iTmp) *(uint32_t *) (hyplnk_base + 0x0c) |= iTmp;

    // Interrupt Status / Clear Register
    iTmp = *(uint32_t *) (hyplnk_base + 0x10);
    if (iTmp) *(uint32_t *) (hyplnk_base + 0x10) |= iTmp;

    // Generate Soft Interrupt Register
    iTmp = *(uint32_t *) (hyplnk_base + 0x18);
    if (iTmp) *(uint32_t *) (hyplnk_base + 0x18) = 0;
}

int mpm_transport_hyplnk_init_peripheral(mpm_transport_hyplnk_t *hypCfg, int index, uint32_t timeout)
{
    hyplnkRet_e retVal;
    int soc_node_id = sl1500_get_nodeid();
    uint32_t dataBase_size = ((hypCfg == 0)? 0x10000000 : 0x08000000);
    uint32_t hyplnk_base, serdes_base;
    int ret=0, loopback, i=0, j = 0, iTmpRevID, iTmp, iTmpECC, iTmpStat, iTmpCtrl, attemptCNT, flag;

    if (hypCfg->hyplnkDir == hyplnk_LOOPBACK) loopback = 1;
    else loopback = 0;

    hyplnkActiveIF[index].physWindowBase = (uint32_t) hyplnkInitCfg.dev.bases[index].dataBase;
    hyplnkInitCfg.dev.bases[index].cfgBase = internal_hyplnk_mmap((uint32_t)(hyplnkInitCfg.dev.bases[index].cfgBase), 0x100, hypCfg->fd[index]);
    hyplnkInitCfg.dev.bases[index].dataBase = internal_hyplnk_mmap((uint32_t)(hyplnkInitCfg.dev.bases[index].dataBase), dataBase_size, hypCfg->fd[index]);
    hyplnkActiveIF[index].windowBase = (uint32_t) hyplnkInitCfg.dev.bases[index].dataBase;
    hyplnkActiveIF[index].configBase = (uint32_t) hyplnkInitCfg.dev.bases[index].cfgBase;

    if ((retVal = Hyplnk_init (&hyplnkInitCfg)) != hyplnk_RET_OK)
    {
       mpm_printf(1,"#[%d]# Hyperlink LLD device configuration failed\n", soc_node_id); fflush(stdout);
       return (-1);
    }

   if (index == 0) {
      serdes_base = (uint32_t) internal_hyplnk_mmap(CSL_HYPERLINK_0_SERDES_CFG_REGS,0x2000, hypCfg->fd[0]);
      hyplnk_base = (uint32_t) internal_hyplnk_mmap(CSL_HYPERLINK_0_SLV_CFG_REGS, 0x100, hypCfg->fd[0]);
   }
   else {
      serdes_base = (uint32_t) internal_hyplnk_mmap(CSL_HYPERLINK_1_SERDES_CFG_REGS,0x2000, hypCfg->fd[1]);
      hyplnk_base = (uint32_t) internal_hyplnk_mmap(CSL_HYPERLINK_1_SLV_CFG_REGS, 0x100, hypCfg->fd[1]);
   }

    // Else try to enable receiver. If no signal detect, exit with -1
    if (hyplnk_enableReceivers(serdes_base, hyplnk_base, loopback, timeout)) { 
        mpm_printf(1, "#[%d]# init_peripheral: error in EnableReceiviers port%d\n", soc_node_id, index);
        timpi_dbg_serdes_dump(soc_node_id, 44, serdes_base, hyplnk_base);
        ret = -1; 
        goto init_end;
    }

    fUserRegFillField32(serdes_base + 0x1fc0 + 0x34,16,2,0);
    usleep(100000);

  // Check for lanes. Retry for 5 times before going to error.
    attemptCNT = 50;
    flag = 0;
    for (j=0; j< attemptCNT; j++) {
        iTmp    = *(uint32_t *) (hyplnk_base + 0x58);
        iTmpECC = *(uint32_t *) (hyplnk_base + 0x4c);
        iTmpRevID = *(uint32_t *) (hyplnk_base + 0x00);
        iTmpStat = *(uint32_t *) (hyplnk_base + 0x08);
        iTmpCtrl = *(uint32_t *) (hyplnk_base + 0x04);
        if (((iTmp & 0xadf0edf0) == 0xadf0adf0) && (iTmpECC == 0) && (flag==0) && ((iTmpCtrl & 0x70)==0)) {
            hyplnk_clear_int(hyplnk_base);
            fUserRegFillField32(hyplnk_base+0x04,0,31,0);
            //fUserRegFillField32(hyplnk_base+0x04,4,11,0);
            //fUserRegFillField32(hyplnk_base+0x04,2,1,0);          //Allow serial operations from portal and remote registers.
            //fUserRegFillField32(hyplnk_base+0x1c,0,4,10);         //TXoverlay (27 significant bits)
            fUserRegFillField32(hyplnk_base+0x1c,0,11,10);
            if ((iTmpStat & 0x180) == 0) break;
        }
        #if 0
        if (((iTmp & 0xadf0edf0) == 0xadf0adf0) && (iTmpECC == 0) && ((iTmpStat & 0x180) == 0)) {
          //Enable serial operations ONLY IF WE HAVE SUCCESFUL LINK UP
          fUserRegFillField32(hyplnk_base+0x04,2,1,0);          //Allow serial operations from portal and remote registers.
          fUserRegFillField32(hyplnk_base+0x1c,0,4,10);         //TXoverlay (27 significant bits)
          break; /* get out of multiple link bring up attempt loop **/
        #endif
        if ( ((iTmp & 0x0000edf0) != 0x0000adf0) || (iTmpECC != 0) || ((iTmpStat & 0x780) != 0) || (flag==1) || ((iTmpCtrl & 0x70)!=0) ) {
            uint32_t i, iAtt, iTxTermNP, iBoost;

            iTxTermNP = termination_workaround1(serdes_base, 1);
            mpm_printf (1, "\n#[%d]# Attempt to bring up link again(%d) port %d, registers (0x00)=%08x (0x58)=%08x (0x4c)=%08x\n", 
            soc_node_id, j, index, iTmpRevID, iTmp, iTmpECC);
            mpm_printf(1, "#[%d]# %d #port%d hyplnkSerdes: iTxTermNP is 0x%x\n", soc_node_id, j, index, iTxTermNP);
            for (i = 0; i < 4; i++) {
            iAtt = iBoost = get_att_boost_values(serdes_base, i, 0);
            iAtt = (iAtt >> 4) & 0x0f;
            iBoost = (iBoost >> 8) & 0x0f;
            mpm_printf(1, "#[%d]# %d #port%d hyplnkSerdes: lane %d; Att = %d, Boost = %d\n", soc_node_id, j, index, i, iAtt, iBoost);
            }
            for (i = 0; i < 4; i ++) {
            mpm_printf (1, "\n#[%d]# %d DLPF lane%d %08x\n", soc_node_id, j, i, get_dlpf (serdes_base, i));
            }

            timpi_dbg_serdes_dump(soc_node_id, j, serdes_base, hyplnk_base);

            *(uint32_t *) (hyplnk_base + 0x4c) = 0; // Reset error counter

            iTmp = *(uint32_t *) (hyplnk_base + 0x58);
//    if ((iTmp&0x0fff0)==0x8df0)  //LPsync but not acknowledged
//    if ( ((iTmp & 0x0000edf0) != 0x0000adf0) || (iTmpECC != 0) || ((iTmpStat & 0x180) != 0))
            {
                usleep(10); //1 uS
                iTmp    = *(uint32_t *) (hyplnk_base + 0x58);
                mpm_printf (1, "\n#[%d]# RECOVERY WORKAROUND - fix RX side, Attempt #%d\n", soc_node_id,j); fflush(stdout);
                fUserRegFillField32(hyplnk_base+0x04,2,1,1);
                while (fSerdesSbRead32(hyplnk_base+0x8)&0x800);
                //if ( (fSerdesSbRead32(hyplnk_base+0x8)&0x100)!=0) fUserRegFillField32(hyplnk_base+0x08,8,1,1);
                fUserRegFillField32 (hyplnk_base+0x04,0,1,1);
                usleep(10);
                fUserRegFillField32 (serdes_base+0x1fc0+0x34,16,1,1); //Set the waterafter SD bit
                for(i=0;i<4;i++) fUserRegFillField32 (serdes_base+0x200+(i*0x200)+0x04,1,2,2); 
                fUserRegFillField32 (hyplnk_base+0x04,0,1,0);
                usleep(10);
                //fUserRegFillField32(hyplnk_base+0x04,4,11,0);
                hyplnk_clear_int(hyplnk_base);
                fUserRegFillField32(hyplnk_base+0x70,16,16,0x0ffff);  //Long Lane Eanble
                fUserRegFillField32(hyplnk_base+0x7C,16,4,0x0f);      //Use Disable State
                fUserRegFillField32(hyplnk_base+0x04,0,31,0);
                if ( (fSerdesSbRead32(hyplnk_base+0x8)&0x100)!=0) fUserRegFillField32(hyplnk_base+0x08,8,1,1);
                if ( (fSerdesSbRead32(hyplnk_base+0x8)&0x80)!=0) fUserRegFillField32(hyplnk_base+0x08,7,1,1);
                usleep(10);
                for(i=0;i<4;i++) fUserRegFillField32 (serdes_base+0x200+(i*0x200)+0x04,1,2,0); //Force Sigdet normal acquire
                while((fSerdesSbRead32(serdes_base+0x1fc0+0x034)&0x0f)!=0x0f);
                for(i = 0; i < 4; i++) boost_workaround_LANE(serdes_base, i); // May need to be moved until signal is up
                fUserRegFillField32 (serdes_base+0x1fc0+0x34,16,1,0); //clear the waterafter SD bit, allow Hyperlink to start receive
                usleep(10); 
                //fUserRegFillField32(hyplnk_base+0x04,2,1,0);
                fUserRegFillField32(hyplnk_base+0x04,0,31,0);
                //fUserRegFillField32(hyplnk_base+0x1c,0,4,10);
                fUserRegFillField32(hyplnk_base+0x1c,0,11,10);
                usleep(10); 
            }
        }
        iTmp    = *(uint32_t *) (hyplnk_base + 0x58);
        iTmpECC = *(uint32_t *) (hyplnk_base + 0x4c);
        iTmpStat = *(uint32_t *) (hyplnk_base + 0x08);
        iTmpCtrl = *(uint32_t *) (hyplnk_base + 0x04);
        if (((iTmp & 0x0000edf0) != 0x0000adf0) || (iTmpECC != 0) || ((iTmpStat & 0x780) != 0) || (iTmpCtrl != 0)) {
            flag = 1;
        }
        else flag = 0;
    } /** end of for ***/

    if (j == attemptCNT) ret = -1;

init_end:
  munmap((void *)serdes_base, 0x2000);
  munmap((void *)hyplnk_base, 0x100);
  
  return ret;
}

int mpm_transport_hyplnk_reset(mpm_transport_hyplnk_t *hypCfg, int index)
{
    hyplnkControlReg_t        control;
    hyplnkStatusReg_t         status;
    hyplnkRegisters_t setRegs;
    memset (&control, 0, sizeof(control));
    memset (&status, 0, sizeof(status));
    memset (&setRegs, 0, sizeof(setRegs));
    setRegs.control        = &control;
    setRegs.status         = &status;
    hyplnkRet_e retVal;
    Hyplnk_Handle handle = NULL;
    if (hypCfg->hyplnkIF == hyplnk_PORT_0)
    {
        hyplnkInitCfg.dev.bases[hypCfg->hyplnkIF].cfgBase = (void *)CSL_HYPERLINK_0_SLV_CFG_REGS;
        hyplnkInitCfg.dev.bases[hypCfg->hyplnkIF].dataBase = (void *)CSL_HYPERLINK_0_SLV_DATA;
    }
    else if (hypCfg->hyplnkIF == hyplnk_PORT_1)
    {
        hyplnkInitCfg.dev.bases[hypCfg->hyplnkIF].cfgBase = (void *)CSL_HYPERLINK_1_SLV_CFG_REGS;
        hyplnkInitCfg.dev.bases[hypCfg->hyplnkIF].dataBase = (void *)CSL_HYPERLINK_1_SLV_DATA;
    }
    return hyplnk_RET_OK;
    if (hypCfg->serdesInit) {

      if ((retVal = Hyplnk_open((int)hypCfg->hyplnkIF, &handle)) != hyplnk_RET_OK) {
          mpm_printf(1, "Open failed\n");
          return retVal;
      }
      memset (&control, 0, sizeof(control));
      do{
            /* Read rev and control */
            if ((retVal = Hyplnk_readRegs (handle, hyplnk_LOCATION_LOCAL, &setRegs)) != hyplnk_RET_OK) {
              mpm_printf(1, "Read revision register failed!\n");
            }
      } while(setRegs.status->rPend);

      control.serialStop = 1;
      if ((retVal = Hyplnk_writeRegs (handle, hyplnk_LOCATION_LOCAL, &setRegs)) != hyplnk_RET_OK) {
          mpm_printf(1, "Control register write failed!\n");
          return retVal;
      }

      memset (&control, 0, sizeof(control));
      do{
            /* Read rev and control */
            if ((retVal = Hyplnk_readRegs (handle, hyplnk_LOCATION_LOCAL, &setRegs)) != hyplnk_RET_OK) {
              mpm_printf(1, "Read revision register failed!\n");
            }
      } while(setRegs.status->rPend);

      //control.serialStop = 0;
      control.reset = 1;
      if ((retVal = Hyplnk_writeRegs (handle, hyplnk_LOCATION_LOCAL, &setRegs)) != hyplnk_RET_OK) {
          mpm_printf(1, "Control register write failed!\n");
          return retVal;
      }
      Hyplnk_close(&handle);
      if ((hyplnkActiveIF[0].slavesActive == 0) && (hyplnkActiveIF[1].slavesActive == 0))
      {
        hyplnkSerdesDisable(hypCfg);
      }
      hyplnkPSCDisable(index);
    }
    return hyplnk_RET_OK;
}

void * mpm_transport_hyplnk_get_window(mpm_transport_hyplnk_t *hypCfg, uint32_t remote_addr, int index, uint32_t *offset)
{
#if TIME_PROFILE_HYPLNK_GET_WINDOW
  struct timespec tp_start, tp_end;
#endif
    hyplnkRet_e retVal;
    Hyplnk_Handle handle = (Hyplnk_Handle) hyplnkActiveIF[(int)hypCfg->hyplnkIF].configBase;
    hyplnkRXAddrSelReg_t  RXAddrSel;
    hyplnkRXPrivIDIdxReg_t RXPrivIDIdx;
    hyplnkRXPrivIDValReg_t RXPrivIDVal;
    hyplnkRXSegIdxReg_t RXSegIdx;
    hyplnkRXSegValReg_t RXSegVal;
    hyplnkRegisters_t     remoteRegs;
    uint32_t globalAddr = (uint32_t)remote_addr;
    uint32_t globalAddrBase, globalAddrOffset;
    uint32_t offsetMask = 0;
    uint32_t reg_sel_shift = 0;
    FILE *fp;
    uint32_t baseAddr;
    int i, iAtt, iBoost;
    int port = (int) hypCfg->hyplnkIF;

    memset (&remoteRegs, 0, sizeof(remoteRegs));
    memset (&RXAddrSel,  0, sizeof(RXAddrSel));
    memset (&RXPrivIDIdx,  0, sizeof(hyplnkRXPrivIDIdxReg_t));
    memset (&RXPrivIDVal,  0, sizeof(hyplnkRXPrivIDValReg_t));
    memset (&RXSegIdx,  0, sizeof(hyplnkRXSegIdxReg_t));
    memset (&RXSegVal,  0, sizeof(hyplnkRXSegValReg_t));

    //localRegs.TXAddrOvly   = &TXAddrOvly;
    remoteRegs.RXAddrSel   = &RXAddrSel;
    remoteRegs.RXPrivIDIdx = &RXPrivIDIdx;
    remoteRegs.RXPrivIDVal = &RXPrivIDVal;
    remoteRegs.RXSegIdx    = &RXSegIdx;
    remoteRegs.RXSegVal    = &RXSegVal;

    RXAddrSel.rxSecHi     = 0;
    RXAddrSel.rxSecLo     = 0;
    RXAddrSel.rxSecSel    = 0;
    RXAddrSel.rxPrivIDSel = hypCfg->hyplnkRxPrivID; /* Symmetric with TXAddrOvly.txPrivIDOvl */
    RXAddrSel.rxSegSel    = hypCfg->hyplnkSegSel;
  
    if (hypCfg->hyplnkSegSel > 0)
      offsetMask = (1<< (hypCfg->hyplnkSegSel+16)) - 1;
    else
      offsetMask = 0xFFFFFFFF;
    
    globalAddrBase   = globalAddr & ~offsetMask;
    if (globalAddrBase == 0)
    {
      globalAddrBase = globalAddr;
      globalAddrOffset = globalAddr & 0xFFFF;
    }
    else
      globalAddrOffset = globalAddr &  offsetMask;

  RXPrivIDIdx.rxPrivIDIdx = index;
  RXSegIdx.rxSegIdx = index;  
  RXPrivIDVal.rxPrivIDVal = index;
  RXSegVal.rxLenVal = hypCfg->hyplnkRxLen;
  RXSegVal.rxSegVal = globalAddrBase >> 16;
#if TIME_PROFILE_HYPLNK_GET_WINDOW
  clock_gettime(CLOCK_MONOTONIC, &tp_start);
#endif
  if (port == 0)
    baseAddr = (uint32_t) internal_hyplnk_mmap(CSL_HYPERLINK_0_SERDES_CFG_REGS,0x2000, hypCfg->fd[0]);
  else if (port == 1)
    baseAddr = (uint32_t) internal_hyplnk_mmap(CSL_HYPERLINK_1_SERDES_CFG_REGS,0x2000, hypCfg->fd[1]);

  munmap((void *)baseAddr, 0x2000);
    if ((retVal = Hyplnk_writeRegs (handle, hyplnk_LOCATION_REMOTE, &remoteRegs)) != hyplnk_RET_OK) {
        mpm_printf(1, "mpm_transport_hyplnk_get_window: Remote register write failed!\n");
        return NULL;
    }
#if TIME_PROFILE_HYPLNK_GET_WINDOW
  clock_gettime(CLOCK_MONOTONIC, &tp_end);
  mpm_printf(1, "[TIME PROFILE: mpm_transport_hyplnk_get_window] Time taken for Hyplnk_writeRegs() to Remote: %lld\n",clock_diff (&tp_start, &tp_end));
#endif

    if (hypCfg->hyplnkSegSel)
    {
        reg_sel_shift = index << (hypCfg->hyplnkSegSel + 16);
    }
    if (reg_sel_shift >= 0x10000000)
    {
      mpm_printf(1, "Index out of bounds, please use smaller segment sizes and a lower rxseg value\n");
      return NULL;
    }
    *offset = globalAddrOffset + reg_sel_shift;

    mpm_printf(2, "index is 0x%08x\n",index);
    mpm_printf(2, "globalAddr is 0x%08x\n", globalAddr);
    mpm_printf(2, "globalAddrBase is 0x%08x\n", globalAddrBase);
    mpm_printf(2, "globalAddrOffset is 0x%08x\n", globalAddrOffset);
    mpm_printf(2, "reg_sel_shift is 0x%08x\n", reg_sel_shift);

    return (void *)(hyplnkActiveIF[(int)hypCfg->hyplnkIF].windowBase);
}

void hyplnkPeripheralSetup(mpm_transport_hyplnk_t *hypCfg) {
  uint32_t serdes_base, hyplnk_base;
  int ret, loopback;
  uint32_t serdes_base0, hyplnk_base0;
  uint32_t serdes_base1, hyplnk_base1;
  int index = (int) hypCfg->hyplnkIF;

  if (hypCfg->hyplnkDir == hyplnk_LOOPBACK) loopback = 1;
  else loopback = 0;

  serdes_base0 = (uint32_t) internal_hyplnk_mmap(CSL_HYPERLINK_0_SERDES_CFG_REGS,0x2000, hypCfg->fd[0]);
  hyplnk_base0 = (uint32_t) internal_hyplnk_mmap(CSL_HYPERLINK_0_SLV_CFG_REGS, 0x100, hypCfg->fd[0]);
  serdes_base1 = (uint32_t) internal_hyplnk_mmap(CSL_HYPERLINK_1_SERDES_CFG_REGS,0x2000, hypCfg->fd[1]);
  hyplnk_base1 = (uint32_t) internal_hyplnk_mmap(CSL_HYPERLINK_1_SLV_CFG_REGS, 0x100, hypCfg->fd[1]);
  /*** Start transmitters ****/
  //hyplnk 0
  if ((uioutil_check_device_exist("hyperlink0") == 1) && (index==0)) {
    ret = hyplnk_startupTX(serdes_base0, hyplnk_base0, loopback, 0);
  }
  //hyplnk 1
  if ((uioutil_check_device_exist("hyperlink1") == 1) && (index==1)) {
    ret = hyplnk_startupTX(serdes_base1, hyplnk_base1, loopback, 1);
  }

  munmap((void *)serdes_base0, 0x2000);
  munmap((void *)hyplnk_base0, 0x100);
  munmap((void *)serdes_base1, 0x2000);
  munmap((void *)hyplnk_base1, 0x100);
}

/************************************************************************************/
int hyplnkPeripheralSerdesHealth(unsigned int iHyperLinkBase, unsigned int in_base)
{
  int i, ret = 0;

  if ((fSerdesSbRead32(iHyperLinkBase+0x58)& 0xadf0edf0) != 0xadf0adf0) ret = -1;
  if (fSerdesSbRead32(iHyperLinkBase+0x4c)!=0) ret = -1;
  if(fSerdesSbRead32(in_base+0x1fc0+0x034) != 0xf0000f0f) ret = -1;

  if(ret < 0) {
    mpm_printf(1, "hyplnkPeripheralSerdesHealth: ret < 0 !!!\n");
    mpm_printf(1, "Line status (%08x)!\n", fSerdesSbRead32(iHyperLinkBase+0x58));
    //Ensure Hyperlink has linked and running
    for (i = 0; i < 4; i ++) {
      mpm_printf (1, "DLPF lane%d %08x\n", i, get_dlpf (in_base, i));
    }
    mpm_printf(1, "ECC counter (%08x)!\n", fSerdesSbRead32(iHyperLinkBase+0x4c));
    mpm_printf(1, "Link status (%08x)!\n", fSerdesSbRead32(in_base+0x1fc0+0x034));
  }
  else {
    mpm_printf(1, "hyplnkPeripheralSerdesHealth return health okay!\n");
  }
  return ret;
}

int hyplnk_startupTX(uint32_t in_base, uint32_t iHyperLinkBase, int loopback, int index)
{
  CSL_SERDES_STATUS             pllstat;
  int i, iLanes=4;
  uint32_t iTmp, iTxTermNP;
  char lanerate[MAX_FILE_NAME_LEN];
  int lane_rate = 5; //half
  CSL_SERDES_REF_CLOCK          refClock = CSL_SERDES_REF_CLOCK_312p5M;
  CSL_SERDES_LINK_RATE          linkRate = CSL_SERDES_LINK_RATE_6p25G;
  uint32_t                      baseAddr, baseAddr1, baseAddr2;
  uint32_t serdes_regs;
  int cm, c1, c2, refclk, serrate, ret;
  char device[MAX_FILE_NAME_LEN];
  int conf_coeff = 1;
  CSL_SERDES_TX_COEFF_T tx_coeff;
  FILE *fp;

  csl_serdes_phy_type phy = SERDES_HYPERLINK;

  if (index == 0) strcpy(device, "hyperlink0");
  else strcpy(device, "hyperlink1");

  ret = uioutil_get_param_int(device, "ti,serdes_cm", &cm);
  if (ret == -1) conf_coeff = 0;
  ret = uioutil_get_param_int(device, "ti,serdes_c1", &c1);
  if (ret == -1) conf_coeff = 0;
  ret = uioutil_get_param_int(device, "ti,serdes_c2", &c2);
  if (ret == -1) conf_coeff = 0;
  ret = uioutil_get_param_int(device, "ti,serdes_refclk_khz", &refclk);
  if (ret == -1) {
    mpm_printf(1, "Error: can't find refclk rate\n");
    return -1;
  }
  ret = uioutil_get_param_int(device, "ti,serdes_maxserrate_khz", &serrate);
  if (ret == -1) {
    mpm_printf(1, "Error: can't find serial rate\n");
    return -1;
  }
  ret = uioutil_get_param_str(device, "ti,serdes_lanerate", lanerate);
  if (ret == -1) {
    mpm_printf(1, "Error: can't find lane rate\n");
    return -1;
  }    

    if (refclk == 312500)
        refClock = CSL_SERDES_REF_CLOCK_312p5M;
    else if (refclk == 156250)
        refClock = CSL_SERDES_REF_CLOCK_156p25M;
    else 
      return -1;

    if (serrate == 6250000)
        linkRate = CSL_SERDES_LINK_RATE_6p25G;
    else if (serrate == 12500000)
        linkRate = CSL_SERDES_LINK_RATE_12p5G;
    else 
      return -1;

    if (strcmp(lanerate,"full") == 0)
        lane_rate = 4;
    else if (strcmp(lanerate,"half") == 0)
        lane_rate = 5;
    else if (strcmp(lanerate,"quarter") == 0)
        lane_rate = 6;
    else 
      return -1;

    mpm_printf (1, "\nstartupTX1: %d %08x ", index, iHyperLinkBase);

#if orig
  fUserRegFillField32(in_base + 0x1fc0 + 0x34,29,3,0x4);  //Disable the PLL
  for(i=0;i<iLanes;i++)  //Disable lanes
    {
    fUserRegFillField32(in_base + 0x1fc0 + 0x20 + (i*4),29,3,4);
    fUserRegFillField32(in_base + 0x1fc0 + 0x20 + (i*4),13,3,4);
    }
#else
  fUserRegFillField32(in_base + 0x1fc0 + 0x34,29,3,0x4);  // DJDBG: Wrongly set to 0 before last change - Disable the PLL
  for(i=0;i<iLanes;i++)  //Disable lanes
  {
    fUserRegFillField32(in_base + 0x1fc0 + 0x20 + (i*4),29,3,4);  //DJDBG: !!!THIS IS NOT RIGHT!!!, use '4' instead, like in #if orig case, We donot want Hyperlink to run right away.
    fUserRegFillField32(in_base + 0x1fc0 + 0x20 + (i*4),13,3,4);
  }
#endif

#if orig
  csl_wiz8_sb_refclk312p5MHz_20bit_12p5Gbps_sr1(in_base);  //Load appropriate config file
#else
  CSL_HyperlinkSerdesInit(in_base, refClock, linkRate);
#endif

  for(i=0;i<iLanes;i++)
    fSbSerdesPhyaPatchLane(in_base,i);   //DJDBG: Apply only to >8Gbps

  mpm_printf (1, "\nstartupTX2: %d %08x ", index, iHyperLinkBase);

#if 0  
 //DJDBG: THERE IS NEW ONE Disable Transmitter. SNOWBUSH suggested new one
 //for(i=0;i<iLanes;i++) fUserRegFillField32(in_base + (i * 0x200) + 0x200 + 0x0b8,16,2,3); //(PhyA) Disable transmitter, Do this before we enable the lanes the first time.
 for(i=0;i<iLanes;i++) fUserRegFillField32(in_base + (i * 0x200) + 0x200 + 0x0b8,16,2,3); //(PhyA) Disable transmitter, Do this before we enable the lanes the first time.
#else
 for(i=0;i<iLanes;i++) fUserRegFillField32(in_base + 0x1FC0 + 0x20 + (i * 0x4),24,2,3); //(PhyA) Force tranmit idle (prevents remote receiver from seeing training patterning and linking falsely
#endif

 for(i=0;i<iLanes;i++) fUserRegFillField32(in_base + (i * 0x200) + 0x200 + 0x04,1,2,2);  //Force SERDES Signal Detect LO (Reset CDR, Att and Boost)

#if orig
  fUserRegFillField32(in_base + 0x0a00 + 0x0c8, 0,6,VUSR_PATTERN_MASK);  //Setup ignore mask for boost pattern
#else
  fUserRegFillField32(in_base + 0x0a00 + 0x0c8, 0,6,0x03); //DJDBG: perhaps 0x0f?
  //fUserRegFillField32(in_base + 0x0a00 + 0x0c8, 0,6,0x0f); //DJDBG: perhaps 0x0f?
#endif

  //Setup cm, c1 and c2 SERDES Parameters
  fSerdesSbAssertReset(in_base);

  mpm_printf (1, "\nstartupTX3: %d %08x ", index, iHyperLinkBase);

#if orig
  if ((in_base==0x0231a000)) {fSerdesSbSetup_cm_c1_c2(in_base,0,h0l0cm, h0l0cm, h0l0cm); if (VUSR_FORCE_ATTBOOST) fSerdesSbPhyASetupAttnBoost(in_base,0, h0l0at, h0l0bo);}
  if ((in_base==0x0231a000)) {fSerdesSbSetup_cm_c1_c2(in_base,1,h0l1cm, h0l1cm, h0l1cm); if (VUSR_FORCE_ATTBOOST) fSerdesSbPhyASetupAttnBoost(in_base,1, h0l1at, h0l1bo);}
  if ((in_base==0x0231a000)) {fSerdesSbSetup_cm_c1_c2(in_base,2,h0l2cm, h0l2cm, h0l2cm); if (VUSR_FORCE_ATTBOOST) fSerdesSbPhyASetupAttnBoost(in_base,2, h0l2at, h0l2bo);}
  if ((in_base==0x0231a000)) {fSerdesSbSetup_cm_c1_c2(in_base,3,h0l3cm, h0l3cm, h0l3cm); if (VUSR_FORCE_ATTBOOST) fSerdesSbPhyASetupAttnBoost(in_base,3, h0l3at, h0l3bo);}
  if ((in_base==0x0231c000)) {fSerdesSbSetup_cm_c1_c2(in_base,0,h1l0cm, h1l0cm, h1l0cm); if (VUSR_FORCE_ATTBOOST) fSerdesSbPhyASetupAttnBoost(in_base,0, h1l0at, h1l0bo);} 
  if ((in_base==0x0231c000)) {fSerdesSbSetup_cm_c1_c2(in_base,1,h1l1cm, h1l1cm, h1l1cm); if (VUSR_FORCE_ATTBOOST) fSerdesSbPhyASetupAttnBoost(in_base,1, h1l1at, h1l1bo);} 
  if ((in_base==0x0231c000)) {fSerdesSbSetup_cm_c1_c2(in_base,2,h1l2cm, h1l2cm, h1l2cm); if (VUSR_FORCE_ATTBOOST) fSerdesSbPhyASetupAttnBoost(in_base,2, h1l2at, h1l2bo);} 
  if ((in_base==0x0231c000)) {fSerdesSbSetup_cm_c1_c2(in_base,3,h1l3cm, h1l3cm, h1l3cm); if (VUSR_FORCE_ATTBOOST) fSerdesSbPhyASetupAttnBoost(in_base,3, h1l3at, h1l3bo);} 
#else
  for (i=0; i<4; i++) {
    CSL_SERDES_CONFIG_CM_C1_C2(in_base, i, cm, c1, c2, 5); //DJDBG: What is 5?
  }
#endif

  fSerdesSbDeassertReset(in_base,0);   //Don’t wait for Lane OK since the Hyperlink is not driving lane enable yet…
  mpm_printf (1, "\nstartupTX4: %d %08x ", index, iHyperLinkBase);
#if orig
  //Setup Lane Rates if needed
  iLaneRate=VUSR_LANE_RATE;              //0,4=Full, 5=Half, 6=Quarter
  for(i=0;i<iLanes;i++)
  {
    fUserRegFillField32(in_base + 0x1fc0 + 0x20 + (i*4),26,3,iLaneRate);
    fUserRegFillField32(in_base + 0x1fc0 + 0x20 + (i*4),10,3,iLaneRate);
  }
  if ((in_base==0x0231a000)) status = enable_module(vusr0_pdctl,vusr0_mdctl);  //Start Hyperlink 0 if requested
  if ((in_base==0x0231c000)) status = enable_module(vusr1_pdctl,vusr1_mdctl);  //Start Hyperlink 1 if requested
  if (status != PASS)
  {
    mpm_printf (1, "\nstartupTX4-PSC: %d %08x ", index, iHyperLinkBase);
    return(FAILED_PSC);
  }
#else
  for(i=0;i<iLanes;i++)
  {
     fUserRegFillField32(in_base + 0x1fc0 + 0x20 + (i*4),26,3,lane_rate);
     fUserRegFillField32(in_base + 0x1fc0 + 0x20 + (i*4),10,3,lane_rate);
  }
  hyplnkPSCSetup(index);
  fUserRegFillField32(iHyperLinkBase+0x04,0,1,0); //After PSC enable, IVAN Oct 2nd
#endif

  mpm_printf (1, "\nstartupTX5: %d %08x ", index, iHyperLinkBase);
  fUserRegFillField32(iHyperLinkBase+0x04,2,1,1);          //Set Serial Stop, lets debugger look at VUSR registers without hanging!
  //Set to one lane (2) or 4 lane (4)
  fUserRegFillField32(iHyperLinkBase+0x44,0,3,4);          //4-four,2-1lane

if (loopback) {
  fUserRegFillField32(iHyperLinkBase+0x04,1,1,1);          //Internal loopback (Set iLoop)
  //Setup SERDES Digital loopback
  for(i=0;i<iLanes;i++) fUserRegFillField32(in_base+(i*0x200)+0x200+0x00,30,1,1);
}

  fUserRegFillField32(iHyperLinkBase+0x70,16,16,0x0ffff);  //Long Lane Eanble
  fUserRegFillField32(iHyperLinkBase+0x7C,16,4,0x0f);      //Use Disable State

  fUserRegFillField32(in_base + 0x1fc0 + 0x34,16,2,3);     //Suspend VUSR Link Up
  for(i=0;i<iLanes;i++)  //Enable Lanes
    {
    fUserRegFillField32(in_base + 0x1fc0 + 0x20 + (i*4),29,3,7);  //If we override the Tx, then Tx IDLE does not work! Let Hyperlink control it!
    fUserRegFillField32(in_base + 0x1fc0 + 0x20 + (i*4),13,3,7);  //Let Hyperlink power down SERDES as well.
    }
  fUserRegFillField32(in_base + 0x1fc0 + 0x34,29,3,0x7);  //Enable the PLL

  mpm_printf (1, "\nstartupTX6: %d %08x ", index, iHyperLinkBase);
#if orig
  //At this time it will take approx 700uS for the PLL to come up
  iTmp=fPoll32(in_base+0x1fc0+0x034,0x10000000,0x10000000,700000ll); //Wait 700uS for PLL OK
  if (iTmp) {printf("PHYB PLLs Not Locked\n"); return(FAILED_PLL_LOCK);}

  //At this tile it will take 400-700uS for the lanes to come up
  iTmp=fPoll32(in_base+0x1fc0+0x034,0x00000f00,0x00000f00,700000ll); //Wait 700uS for LANE OK
  if (iTmp) {printf("PHYB Lanes are not OK\n"); return(FAILED_LANE_OK);}
#else
  mpm_printf (1, "\nstartupTX7: %d %08x ", index, iHyperLinkBase);
  do {
    pllstat = CSL_HyperlinkSerdesGetStatus(in_base, 4);
  } while(pllstat == CSL_SERDES_STATUS_PLL_NOT_LOCKED);
  mpm_printf (1, "\nstartupTX8: %d %08x\n", index, iHyperLinkBase);
#endif

  iTxTermNP = termination_workaround1(in_base, 1);
  termination_workaround2(in_base, 4, iTxTermNP);

  mpm_printf (1, "\nstartupTX9: %d %08x ", index, iHyperLinkBase);
  //Since at this time the transmitter is disabled we could read the termination values after the lane 0 is OK and apply it to the all lanes. 
  //But is we wait for all lanes to be OK, then the termination will be applied from lane 0 to the other three lanes.
  
#if orig
  // Offset averaging workaround
  fSerdesSbFindAverageOffsets(in_base,iLanes,100);
  fSerdesSbWriteAverageOffsets(in_base,iLanes);
#else  
  tx_coeff.cm_coeff   = cm;
  tx_coeff.c1_coeff   = c1;
  tx_coeff.c2_coeff   = c2;
  tx_coeff.tx_att     = 12;
  tx_coeff.tx_vreg    = 4;
  CSL_Serdes_DFE_OffsetCalibration(in_base, 4, SERDES_HYPERLINK, tx_coeff);
#endif

  //Now enable the transmitter since tx term and cm, c1 and c2 are all valid!

  mpm_printf (1, "\nstartupTX10: %d %08x ", index, iHyperLinkBase);
  //for(i=0;i<iLanes;i++) fUserRegFillField32(in_base + (i * 0x200) + 0x200 + 0x0b8,16,2,0); //(PhyA) Enable transmitter
  for(i=0;i<iLanes;i++) fUserRegFillField32(in_base + 0x1FC0 + 0x20 + (i * 0x4),24,2,0); //(PhyA) Disable transmit idle (allow receiver to starup)
//===============================================================
//BREAK FUNCTION AT THIS POINT, and create another for Receivier. 
//At this point Transmitters are transmitting TRAINING DATA
//===============================================================
  mpm_printf (1, "\nstartupTX11: %d %08x\n", index, iHyperLinkBase);
  //fclose(fp);
}

int hyplnk_enableReceivers(uint32_t in_base, uint32_t iHyperLinkBase, int loopback, int timeout)
{
  int i=0, iLanes=4;
  uint32_t iTmp;
  int ret = 0;
  int soc_node_id = sl1500_get_nodeid();

  //Since the receiver is still held off this is where we will wait for the other side to start, 
  mpm_printf (1, "\n#[%d]#startupRX1: %08x\n", soc_node_id, iHyperLinkBase);
  //Allow receiver to start
 for(i=0;i<iLanes;i++) fUserRegFillField32(in_base + (i * 0x200) + 0x200 + 0x04,1,2,loopback ? 3 : 0);  //Allow SERDES to reacquire data

  if (timeout) {
    i=0;
    mpm_printf (1, "\n#[%d]#Signal Detect timeout: %d\ni=", soc_node_id, timeout);
    while (i<timeout) {
      iTmp=fSerdesSbRead32(in_base+0x1fc0+0x034)&0x0f; 
      if (iTmp == 0xf) break;
      usleep(1000);
      i++;
    }
    if (i==timeout) {
      mpm_printf(1, "\n#[%d]# init_peripheral: Timeout occurred waiting for signal detect! timeout = %d\n", soc_node_id, timeout);
      for(i=0;i<iLanes;i++) fUserRegFillField32(in_base + (i * 0x200) + 0x200 + 0x04,1,2,2);  //Force SERDES Signal Detect LO (Reset CDR, Att and Boost)
      ret = -1;
    }
  }
  else {
    do
    {
      iTmp=fSerdesSbRead32(in_base+0x1fc0+0x034)&0x0f;  // check if all 4 lanes are up - remote side
    } while(iTmp!=0x0f);    //Stay here until the other side reached the above transmit enable.
  }

  mpm_printf (1, "\n#[%d]#enableReceivers DONE: %08x\n", soc_node_id, iHyperLinkBase);
  return ret;
}

/********************************************************************************************/
int hyplnk_startupRX(uint32_t in_base, uint32_t iHyperLinkBase, int loopback)
{
  CSL_SERDES_STATUS             pllstat;
  int i, iLanes=4;
  uint32_t iTmp, iTxTermNP;
  char lanerate[MAX_FILE_NAME_LEN];
  int lane_rate = 5; //half
  CSL_SERDES_REF_CLOCK          refClock = CSL_SERDES_REF_CLOCK_312p5M;
  CSL_SERDES_LINK_RATE          linkRate = CSL_SERDES_LINK_RATE_6p25G;
  uint32_t                      baseAddr, baseAddr1, baseAddr2;
  uint32_t serdes_regs;
  int cm, c1, c2, refclk, serrate, ret;
  char device[MAX_FILE_NAME_LEN];
  int conf_coeff = 1;
  FILE *fp;
  int soc_node_id = sl1500_get_nodeid();
  int dd;

  csl_serdes_phy_type phy = SERDES_HYPERLINK;

#if orig
  //Since we have RxValid do the attenuation/boost fix 
  for(i=0;i<iLanes;i++) fSerdesAttenuationBoostWoraround(in_base,i);
#else
  mpm_printf (1, "\n#[%d]# startupRX4: %08x ", soc_node_id, iHyperLinkBase);

//  if (!boost_workaround(baseAddr, 4, 0)) { //We poll checkrx_valid
//    fprintf(fp, "\nstartupRX4: boost workaround failed %08x\n", iHyperLinkBase);
//  }
    for(i = 0; i < iLanes; i++)
    {
      boost_workaround_LANE(in_base, i); // May need to be moved until signal is up
    }
  mpm_printf (1, "\n#[%d]# startupRX5: %08x ", soc_node_id, iHyperLinkBase);
#endif
  do
  {
    iTmp=fSerdesSbRead32(in_base+0x1fc0+0x034)&0x0f;
  } while(iTmp!=0x0f);    //Stay here until the other side reached the above transmit enable.

  mpm_printf (1, "\n#[%d]# startupRX6: %08x ", soc_node_id, iHyperLinkBase);
  //Now allow Hyperlink to startup (release MAC)
  fUserRegFillField32(in_base + 0x1fc0 + 0x34,16,2,0); //Resume VUSR Link Up ???

#if orig
  fCycleDelay(1000000); Wait for Hyperlink start
#else
  usleep(1000000); //1 second of delay
#endif

#if 1 //Based on 09-18 suggestion

#if orig
  if (!fCheckReg32(iHyperLinkBase+0x58,8,1,1)) {
    return(FAILED_HYPERSTART_NO_LINK);  //Ensure Hyperlink has linked and running
  }

  if (fRead(iHyperLinkBase+0x4c)!=0) return(FAILED_HYPERLINK_UNSTABLE);
#else

  /**********************************************************************/
  if ((fSerdesSbRead32(iHyperLinkBase+0x58)& 0xadf0edf0) != 0xadf0adf0)
  {
    //Ensure Hyperlink has linked and running (BOTH DIRECTIONS!!)
    mpm_printf (1, "\n#[%d]# startupRX6err1: %08x (reg(0x58)=%08x\n", soc_node_id,
       iHyperLinkBase, fSerdesSbRead32(iHyperLinkBase+0x58));
    for (i = 0; i < 16; i ++) {
       iTmp=fSerdesSbRead32(in_base+0x1fc0+i * 4);
       mpm_printf (1, "[%08x] %08x\n", 0x1fc0 + i * 4, iTmp);
    }
    for (i = 0; i < iLanes; i ++) {
      mpm_printf (1, "DLPF lane%d %08x\n", i, get_dlpf (in_base, i));
    }
  } 

  if (fSerdesSbRead32(iHyperLinkBase+0x4c)!=0) {
    mpm_printf (1, "\n#[%d]# startupRX6err2: %08x\n", soc_node_id, iHyperLinkBase);
  } 
#endif
  mpm_printf (1, "\n#[%d]# startupRX7: %08x\n", soc_node_id, iHyperLinkBase);

//  fUserRegFillField32(iHyperLinkBase+0x04,2,1,0);          //Allow serial operations from portal and remote registers.
//  fUserRegFillField32(iHyperLinkBase+0x1c,0,4,10);         //TXoverlay (27 significant bits)
#endif
  return 0;
}
/*********************************************/
